@@ -1,7 +1,7 @@
 /*
  * bibutils.h
  *
- * Copyright (c) Chris Putnam 2005-2014
+ * Copyright (c) Chris Putnam 2005-2020
  *
  * Source code released under GPL version 2
  *
@@ -14,15 +14,11 @@ extern "C" {
 #endif /* __cplusplus */
 
 #include <stdio.h>
+#include "bibdefs.h"
 #include "bibl.h"
-#include "list.h"
+#include "slist.h"
 #include "charsets.h"
-#include "newstr_conv.h"
-
-#define BIBL_OK           (0)
-#define BIBL_ERR_BADINPUT (-1)
-#define BIBL_ERR_MEMERR   (-2)
-#define BIBL_ERR_CANTOPEN (-3)
+#include "str_conv.h"
 
 #define BIBL_FIRSTIN      (100)
 #define BIBL_MODSIN       (BIBL_FIRSTIN)
@@ -36,7 +32,8 @@ extern "C" {
 #define BIBL_BIBLATEXIN   (BIBL_FIRSTIN+8)
 #define BIBL_EBIIN        (BIBL_FIRSTIN+9)
 #define BIBL_WORDIN       (BIBL_FIRSTIN+10)
-#define BIBL_LASTIN       (BIBL_FIRSTIN+10)
+#define BIBL_NBIBIN       (BIBL_FIRSTIN+11)
+#define BIBL_LASTIN       (BIBL_FIRSTIN+11)
 
 #define BIBL_FIRSTOUT     (200)
 #define BIBL_MODSOUT      (BIBL_FIRSTOUT)
@@ -46,9 +43,20 @@ extern "C" {
 #define BIBL_ISIOUT       (BIBL_FIRSTOUT+4)
 #define BIBL_WORD2007OUT  (BIBL_FIRSTOUT+5)
 #define BIBL_ADSABSOUT    (BIBL_FIRSTOUT+6)
-#define BIBL_LASTOUT      (BIBL_FIRSTOUT+6)
+#define BIBL_NBIBOUT      (BIBL_FIRSTOUT+7)
+#define BIBL_BIBLATEXOUT  (BIBL_FIRSTOUT+8)
+#define BIBL_LASTOUT      (BIBL_FIRSTOUT+8)
 
-#define BIBL_FORMAT_VERBOSE (1)
+#define BIBL_FORMAT_VERBOSE             (1)
+#define BIBL_FORMAT_BIBOUT_FINALCOMMA   (2)
+#define BIBL_FORMAT_BIBOUT_SINGLEDASH   (4)
+#define BIBL_FORMAT_BIBOUT_WHITESPACE   (8)
+#define BIBL_FORMAT_BIBOUT_BRACKETS    (16)
+#define BIBL_FORMAT_BIBOUT_UPPERCASE   (32)
+#define BIBL_FORMAT_BIBOUT_STRICTKEY   (64)
+#define BIBL_FORMAT_BIBOUT_SHORTTITLE (128)
+#define BIBL_FORMAT_BIBOUT_DROPKEY    (256)
+#define BIBL_FORMAT_MODSOUT_DROPKEY   (512)
 
 #define BIBL_RAW_WITHCHARCONVERT (4)
 #define BIBL_RAW_WITHMAKEREFID   (8)
@@ -64,9 +72,9 @@ extern "C" {
 #define BIBL_SRC_FILE    (1)  /* value from file, priority over default */
 #define BIBL_SRC_USER    (2)  /* value from user, priority over file, default */
 
-#define BIBL_XMLOUT_FALSE    NEWSTR_CONV_XMLOUT_FALSE
-#define BIBL_XMLOUT_TRUE     NEWSTR_CONV_XMLOUT_TRUE
-#define BIBL_XMLOUT_ENTITIES NEWSTR_CONV_XMLOUT_ENTITIES
+#define BIBL_XMLOUT_FALSE    STR_CONV_XMLOUT_FALSE
+#define BIBL_XMLOUT_TRUE     STR_CONV_XMLOUT_TRUE
+#define BIBL_XMLOUT_ENTITIES STR_CONV_XMLOUT_ENTITIES
 
 typedef unsigned char uchar;
 
@@ -95,36 +103,36 @@ typedef struct param {
 	uchar verbose;
 	uchar singlerefperfile;
 
-	list asis;  /* Names that shouldn't be mangled */
-	list corps; /* Names that shouldn't be mangled-MODS corporation type */
+	slist asis;  /* Names that shouldn't be mangled */
+	slist corps; /* Names that shouldn't be mangled-MODS corporation type */
 
 	char *progname;
 
 
-        int  (*readf)(FILE*,char*,int,int*,newstr*,newstr*,int*);
-        int  (*processf)(fields*,char*,char*,long);
+        int  (*readf)(FILE*,char*,int,int*,str*,str*,int*);
+        int  (*processf)(fields*,const char*,const char*,long,struct param*);
         int  (*cleanf)(bibl*,struct param*);
-        int  (*typef) (fields*,char*,int,struct param*,variants*,int);
-        int  (*convertf)(fields*,fields*,int,struct param*,variants*,int);
+        int  (*typef) (fields*,const char*,int,struct param*);
+        int  (*convertf)(fields*,fields*,int,struct param*);
         void (*headerf)(FILE*,struct param*);
         void (*footerf)(FILE*);
-        void (*writef)(fields*,FILE*,struct param*,unsigned long);
+	int  (*assemblef)(fields*,fields*,struct param*,unsigned long);
+        int  (*writef)(fields*,FILE*,struct param*,unsigned long);
         variants *all;
         int  nall;
 
 
 } param;
 
-extern void bibl_initparams( param *p, int readmode, int writemode,
-	char *progname );
-extern void bibl_freeparams( param *p );
-extern int  bibl_readasis( param *p, char *filename );
-extern int  bibl_addtoasis( param *p, char *entry );
-extern int  bibl_readcorps( param *p, char *filename );
-extern int  bibl_addtocorps( param *p, char *entry );
-extern int  bibl_read( bibl *b, FILE *fp, char *filename, param *p );
-extern int  bibl_write( bibl *b, FILE *fp, param *p );
-extern void bibl_reporterr( int err );
+int  bibl_initparams( param *p, int readmode, int writemode, char *progname );
+void bibl_freeparams( param *p );
+int  bibl_readasis( param *p, char *filename );
+int  bibl_addtoasis( param *p, char *entry );
+int  bibl_readcorps( param *p, char *filename );
+int  bibl_addtocorps( param *p, char *entry );
+int  bibl_read( bibl *b, FILE *fp, char *filename, param *p );
+int  bibl_write( bibl *b, FILE *fp, param *p );
+void bibl_reporterr( int err );
 
 #ifdef __cplusplus
 }
